# Author: Anthony Stahler
# Last updated: Apr 10, 2022
# License: Free for personal use, for commercial use please reach out to the author via the BestEdit contact page: http://bestedit.net/contact/
# The most recent version of this script can be found at http://bestedit.net/premiere-pro-xml-fix/
#
# Version 2: added the ability to shift each file by a different number of frames, if desired
# Version 3: Correct instruction text
# Version 4: Fixed a bunch of bugs with the fta_ht functionality - note this script can now generate negative start values which is incorrect but seems to serve as a spacer when loaded in premiere so no reason to truncate to zero.
#
# PLEASE NOTE: For simplicity this script assumes your XML file is 24fps. If it is not, this adjustment script will not work correctly!
# You can use the Premiere Pro XML Fix script to generate XML files that are 24fps.
#
# This script does not normalize the file names or do any of the other fix functionality found in the main XML Fix Script.
# Depending on the length of the source file you may also want to re-trim the very last clip in your project if it has too much footage or dead space at the end compared to what the original project intended
#

# --------------------------
# VARIABLES 
# --------------------------

# this path has subfolders e.g. M:\movie-edits-exports\framerate-shift\output that contain xml files to be processed, this script will create these and the subfolders if you run it and they don't exist
$basepath = "M:\movie-edits-exports\framerate-shift\"

# set this value to your desired amount: 24 will shift it one second forward, -24 will shift one second back.  This will adjust all the ticks appropriately as well.
[long]$frames_to_adjust = 0

# if you leave frames_to_adjust set to zero the script will use this hashtable instead, which allows you to adjust different files by a different amount of frames (for projects with multiple files that are calibrated differently)
# to easily get a list of index for each file, use the list-project-file-indexes powershell script provided at http://bestedit.net/premiere-pro-xml-fix/
# the format is: "IndexNumber" =  FramesToShift;

# commenting out the default example values so we can warn the user if they forgot to put in a frames value
$fta_ht = @{
#"141" = 3; 
#"142" = 0;
#"143" = -5
}

# --------------------------
# CONSTANTS (don't touch)
# --------------------------
$targetext = "*.xml" 

# constant for 24fps 
[long]$pproconst = 10584000000
$fixclipconst = 1
$fixtransconst = 2

# --------------------------
# --------------------------


$srcpath = $basepath 
$destpath = $srcpath + "output\"
$finpath = $srcpath + "done\"

$silenceoutput = New-Item -ItemType Directory -Force -Path $destpath
$silenceoutput = New-Item -ItemType Directory -Force -Path $finpath

# precalculate the ticks to save time
$ticks_to_adjust = $frames_to_adjust * $pproconst

$fta_ticks_ht = @{}

# precalculate the hashtable ticks to save time
foreach ($tmppair in $fta_ht.GetEnumerator())
{
    [long]$tmpcalc = $tmppair.Value * $pproconst

    $fta_ticks_ht.Add($tmppair.Name, $tmpcalc)
}


Get-ChildItem -path ( $srcpath + $targetext ) | foreach {
    $origname1 = $_.Name
    $origname1NOext = $_.BaseName
    $origext = $_.Extension

    $newname = $destpath + $origname1
    $donotmove = $true

    if ($frames_to_adjust -ne 0 -or $fta_ht.Count -gt 0)
    {
        Write-Output "Processing '$origname1NOext'!"

        $outstream = New-Object IO.StreamWriter $newname
        
        $currentfixtype = -1
        [long]$masterclip_id = -1
        [long]$linecount = -1

        [System.IO.File]::ReadLines($srcpath + $origname1) | ForEach-Object {
            $linedata_orig = $_
            $linecount++

            [long]$tmplng = 0
            [long]$tmplng2 = 0

            switch -regex ($linedata_orig)
            {
                '(<clipitem )' { 
                    $currentfixtype = $fixclipconst
                    $masterclip_id = -1
                }

                '(</clipitem>)' { 
                    $currentfixtype = -1
                    $masterclip_id = -1
                }

                '(<transitionitem>)' { 
                    $currentfixtype = $fixtransconst
                }

                '(</transitionitem>)' { 
                    $currentfixtype = -1
                }

                '(<masterclipid>)' { 
                    $tmpstr = $linedata_orig -replace '(<masterclipid>)(.*)(</masterclipid>)', "`$2"
                    $tmpstr = $tmpstr.ToLower().Trim()

                    # make the result just the number
                    $masterclip_id = $tmpstr.Replace("masterclip-", "")
                }

                '(<duration>)(.*)(</duration>)' { 
                    # we want to alter the duration slightly even if it isn't correct for the source, so that no Premiere bugs cause some of the edits to get cut off by duration values
                    # not sure if it ever matters, but better safe than sorry


                    if ($linecount -gt 10) # there is one duration tag at the start of the file that doesn't need to be adjusted
                    {

                        # only bother with it if the new source is longer than the original source
                        $tmplng = ($linedata_orig -replace '(<duration>)(.*)(</duration>)', "`$2").Trim()

                        if ($frames_to_adjust -gt 0)
                        {
                            $tmplng += $frames_to_adjust # duration is the only one with +=, the rest are -= to make the user's shift directions more logical with negative numbers shifting frames left, positive right
                        }
                        else
                        {
                            # we're using the hashtable not the universal adjustment value
                            if ($masterclip_id -ne -1)
                            {
                                if ($fta_ht.ContainsKey("$masterclip_id") -eq $true) 
                                {
                                    $tmplng2 = $fta_ht["$masterclip_id"]

                                    if ($tmplng2 -gt 0) # once again, not trying to be precise, just making sure to only adjust up since this is just making sure the clips are as long or longer than we need, not too short
                                    {
                                        $tmplng += $tmplng2
                                    }
                                }
                                else
                                {
                                    Write-Output "Error1: fta_ht did not contain a key matching '$masterclip_id'!"
                                }
                            }
                            else
                            {

                                Write-Output "Error: missed a frame adjustment, no masterclip_id was set prior to line $linecount!"
                            
                            }
                        }

                        $linedata_orig = $linedata_orig -replace '(?<num1><duration>)(.*)(</duration>)', "`${num1}$tmplng`$2"
                    }
                }

                '(<start>)(.*)(</start>)' { 
                    # don't need to adjust start
                }

                '(<end>)(.*)(</end>)' { 
                    # don't need to adjust end
                }

                '(<cutPointTicks>)(.*)(</cutPointTicks>)' { 
                    # technically this is ticks but since we aren't adjusting framerate we shouldn't need to touch it, since it's an offset value, not an absolute value
                }

                '(<in>)(.*)(</in>)' { 
                    if ($currentfixtype -eq $fixclipconst)
                    {
                        $tmplng = ($linedata_orig -replace '(<in>)(.*)(</in>)', "`$2").Trim()
                        
                        if ($frames_to_adjust -ne 0)
                        {
                            $tmplng -= $frames_to_adjust
                        }
                        else
                        {
                            # we're using the hashtable not the universal adjustment value
                            if ($masterclip_id -ne -1)
                            {
                                if ($fta_ht.ContainsKey("$masterclip_id") -eq $true)
                                {
                                    $tmplng -= $fta_ht["$masterclip_id"]
                                }
                                else
                                {
                                    Write-Output "Error2: fta_ht did not contain a key matching '$masterclip_id'!"
                                }
                            }
                            else
                            {
                                Write-Output "Error: missed a frame adjustment, no masterclip_id was set prior to line $linecount!"
                            }
                        }

                        $linedata_orig = $linedata_orig -replace '(?<num1><in>)(.*)(</in>)', "`${num1}$tmplng`$2"
                    }

                    $donotmove = $false
                }

                '(<out>)(.*)(</out>)' { 
                    if ($currentfixtype -eq $fixclipconst)
                    {
                        $tmplng = ($linedata_orig -replace '(<out>)(.*)(</out>)', "`$2").Trim()
                        
                        if ($frames_to_adjust -ne 0)
                        {
                            $tmplng -= $frames_to_adjust
                        }
                        else
                        {
                            # we're using the hashtable not the universal adjustment value
                            if ($masterclip_id -ne -1)
                            {
                                if ($fta_ht.ContainsKey("$masterclip_id") -eq $true)
                                {
                                    $tmplng -= $fta_ht["$masterclip_id"]
                                }
                                else
                                {
                                    Write-Output "Error3: fta_ht did not contain a key matching '$masterclip_id'!"
                                }
                            }
                            else
                            {
                                Write-Output "Error: missed a frame adjustment, no masterclip_id was set prior to line $linecount!"
                            }
                        }

                        $linedata_orig = $linedata_orig -replace '(?<num1><out>)(.*)(</out>)', "`${num1}$tmplng`$2"
                    }
                }

                '(<pproTicksIn>)(.*)(</pproTicksIn>)' { 
                    if ($currentfixtype -eq $fixclipconst)
                    {
                        $tmplng = ($linedata_orig -replace '(<pproTicksIn>)(.*)(</pproTicksIn>)', "`$2").Trim()
                        
                        if ($frames_to_adjust -ne 0)
                        {
                            $tmplng -= $ticks_to_adjust
                        }
                        else
                        {
                            # we're using the hashtable not the universal adjustment value
                            if ($masterclip_id -ne -1)
                            {
                                if ($fta_ticks_ht.ContainsKey("$masterclip_id") -eq $true)
                                {
                                    $tmplng -= $fta_ticks_ht["$masterclip_id"]
                                }
                                else
                                {
                                    Write-Output "Error4: fta_ticks_ht did not contain a key matching '$masterclip_id'!"
                                }
                            }
                            else
                            {
                                Write-Output "Error: missed a frame adjustment, no masterclip_id was set prior to line $linecount!"
                            }
                        }

                        $linedata_orig = $linedata_orig -replace '(?<num1><pproTicksIn>)(.*)(</pproTicksIn>)', "`${num1}$tmplng`$2"
                    }
                }

                '(<pproTicksOut>)(.*)(</pproTicksOut>)' { 
                    if ($currentfixtype -eq $fixclipconst)
                    {
                        $tmplng = ($linedata_orig -replace '(<pproTicksOut>)(.*)(</pproTicksOut>)', "`$2").Trim()
                        
                        if ($frames_to_adjust -ne 0)
                        {
                            $tmplng -= $ticks_to_adjust
                        }
                        else
                        {
                            # we're using the hashtable not the universal adjustment value
                            if ($masterclip_id -ne -1)
                            {
                                if ($fta_ticks_ht.ContainsKey("$masterclip_id") -eq $true)
                                {
                                    $tmplng -= $fta_ticks_ht["$masterclip_id"]
                                }
                                else
                                {
                                    Write-Output "Error5: fta_ht did not contain a key matching '$masterclip_id'!"
                                }
                            }
                            else
                            {
                                Write-Output "Error: missed a frame adjustment, no masterclip_id was set prior to line $linecount!"
                            }
                        }

                        $linedata_orig = $linedata_orig -replace '(?<num1><pproTicksOut>)(.*)(</pproTicksOut>)', "`${num1}$tmplng`$2"   
                    }    
                }
            }


            $outstream.WriteLine($linedata_orig)
        }

        $outstream.Close()
            
        
        if ($donotmove -eq $false)
        {
            # current time with a bit of randomness to avoid getting the same number over and over since we process fast
            $currenttime = [math]::Floor( (New-TimeSpan -Start (Get-Date "01/01/1970") -End (Get-Date)).TotalSeconds ) - (Get-Random) 

            # fix negative numbers
            if ($currenttime -lt 0)
            {
                $currenttime = -$currenttime
            }

            # zip the created file and delete it
            Compress-Archive -Path $newname -DestinationPath ($destpath + $origname1NOext + "_" + $currenttime + ".zip")
            Remove-Item -LiteralPath ($newname) -Force -Recurse
        }
        else
        {
            Remove-Item -LiteralPath ($newname) -Force -Recurse
        }
    }
    else
    {
        Write-Output "Can't adjust files by 0 frames, please set the `$frames_to_adjust variable to a positive or negative number to shift your project!"
    }

    if ($donotmove -eq $false)
    {
        Get-Item -LiteralPath ( $srcpath + $origname1 ) | Move-Item -Force -Destination { $finpath + $origname1 }
    }
    else
    {
        Write-Output "Found nothing to adjust in '$origname1'!"
    }
}
